/*
  +----------------------------------------------------------------------+
  | PHP Version 5                                                        |
  +----------------------------------------------------------------------+
  | Copyright (c) 1997-2004 The PHP Group                                |
  +----------------------------------------------------------------------+
  | This source file is subject to version 3.0 of the PHP license,       |
  | that is bundled with this package in the file LICENSE, and is        |
  | available at through the world-wide-web at                           |
  | http://www.php.net/license/3_0.txt.                                  |
  | If you did not receive a copy of the PHP license and are unable to   |
  | obtain it through the world-wide-web, please send a note to          |
  | license@php.net so we can mail you a copy immediately.               |
  +----------------------------------------------------------------------+
  | Author: Pierre-Alain Joye <paj@pearfr.org                            |
  +----------------------------------------------------------------------+

  $Id: t_date.c,v 1.7 2004/03/05 13:08:45 pajoye Exp $
*/

#ifdef PHP_WIN32
# include "win32/unistd.h"
#else
# include <unistd.h>
#endif
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#ifndef PHP_WIN32
#include <strings.h>
#endif
#include "php.h"
#include "t_date.h"

/*
 * Create, init, reset, destroy
 */
Date *date_new(DateDay weekstart)
{
	Date *d;
	d = emalloc(sizeof(Date));
	d->weekstart = weekstart;
	d->dmy = d->julian = DATE_FALSE;
	d->status = DATE_TRUE;
	return d;
}

void date_free( Date *d )
{
	efree(d);
}

void date_clear(Date *d, int count)
{
	date_return_if_failed(count != 0);
	date_return_if_failed(d != NULL);
	memset(d, 0x0, count * sizeof(Date));
}

Date *date_new_julian(date_uint julian, DateDay weekstart TSRMLS_DC)
{
	Date *d;
	if (!date_valid_julian(julian)) {
		php_error_docref(NULL TSRMLS_CC, E_WARNING, "Invalid julian day");
		return NULL;
	}
	d = date_new(weekstart);
	if (d) {
		d->julian = DATE_TRUE;
		d->dmy = DATE_FALSE;
		d->julian_days = julian;
		return d;
	}
	return NULL;
}

Date *date_new_dmy(DateDay d, DateMonth m, DateYear y, DateDay weekstart)
{
	Date *newd;
	date_return_val_if_failed(date_valid_dmy(d,m,y),NULL);
	newd = date_new(weekstart);
	if (newd) {
		newd->julian	= DATE_FALSE;
		newd->dmy	= DATE_TRUE;
		newd->day	= d;
		newd->month	= m;
		newd->year	= y;
		return newd;
	}
	return NULL;
}

Date *date_new_now(DateDay weekstart TSRMLS_DC)
{
	Date *newd;
	struct tm *ta, tmbuf;
	time_t timestamp;

	timestamp = time(NULL);
	ta = php_localtime_r(&timestamp, &tmbuf);
	if (!ta) {
		php_error_docref(NULL TSRMLS_CC, E_WARNING, "Cannot perform date calculation");
		return NULL;
	}

	newd = date_new(weekstart);
	if (newd) {
		newd->day	= ta->tm_mday;
		newd->month	= ta->tm_mon+1;
		newd->year	= ta->tm_year+1900;
		newd->hours	= ta->tm_hour;
		newd->minutes	= ta->tm_min;
		newd->seconds	= ta->tm_sec;
		newd->mseconds	= 0;
		newd->dmy	= DATE_TRUE;
		return newd;
	} else {
		php_error_docref(NULL TSRMLS_CC, E_WARNING, "Cannot create a date_time ressource");
		return NULL;
	}
}

DateBoolean date_set_weekstart(Date *d, int ws)
{
	d->weekstart = ws ? DATE_STARTWEEK_MONDAY : DATE_STARTWEEK_SUNDAY;
	return d->weekstart;
}

/*
 * validation
 */

DateBoolean date_valid(const Date *d)
{
	date_return_val_if_failed(d != NULL, DATE_FALSE);
	return d->status && (d->julian || d->dmy);
}

DateBoolean date_valid_julian(int j)
{
	return (j >= DATE_JULIAN_MIN  && j <= DATE_JULIAN_MAX );
}

DateBoolean date_valid_dmy(DateDay d, DateMonth m, DateYear y)
{
	/* ??? DATE_BAD_YEAR==0,
	 * y>DATE_BAD_YEAR is true even if y<
	 */
	return ( (m > DATE_BAD_MONTH) &&
		(m < 13)               &&
		(d > DATE_BAD_DAY)   &&
		(y > DATE_BAD_YEAR)  &&
		(y > 0) &&
		(d <= (date_is_leap_year(y) ?
		days_in_months[1][m] : days_in_months[0][m]) ) );
}

DateBoolean date_valid_day(DateDay d)
{
	return ( (d > DATE_BAD_DAY) && (d < 32) );
}

DateBoolean date_valid_month(DateMonth m)
{
	return ( (m > DATE_BAD_MONTH) && (m < 13) );
}

DateBoolean date_valid_year(DateYear y)
{
	return ( y > DATE_BAD_YEAR );
}

DateBoolean date_valid_weekday(DateWeekday wd)
{
	 return ( (wd > DATE_BAD_WEEKDAY) && (wd < 8) );
}

DateBoolean date_is_leap_year(DateYear year)
{
	date_return_val_if_failed(date_valid_year(year), DATE_FALSE);
	return ((year % 400) == 0 || (((year % 4) == 0) && ((year % 100) != 0)));
}

/*
 * Updates fonction
 *
 */

void date_update_julian(const Date *const_d)
{
	Date *d = (Date *) const_d;
	DateYear year;
	int index;

	date_return_if_failed(d != NULL);
	date_return_if_failed(d->dmy);
	date_return_if_failed(!d->julian);
	date_return_if_failed(date_valid_dmy(d->day, d->month, d->year));

	/* What we actually do is: multiply years * 365 days in the year,
	*  add the number of years divided by 4, subtract the number of
	*  years divided by 100 and add the number of years divided by 400,
	*  which accounts for leap year stuff. Code from Steffen Beyer's
	*  DateCalc.
	*/

	/* we know d->year > 0 since it's valid */
	year = d->year - 1;

	d->julian_days = year * 365U;
	d->julian_days += (year >>= 2); /* divide by 4 and add */
	d->julian_days -= (year /= 25); /* divides original # years by 100 */
	d->julian_days += year >> 2;    /* divides by 4, which divides original by 400 */

	index = date_is_leap_year(d->year) ? 1 : 0;

	d->julian_days += days_in_year[index][d->month] + d->day;

	date_return_if_failed(date_valid_julian(d->julian_days));

	d->julian = DATE_TRUE;
}

void date_update_dmy(const Date *const_d)
{
	Date *d = (Date *) const_d;
	DateYear y;
	DateMonth m;
	DateDay day;

	date_uint A, B, C, D, E, M;

	date_return_if_failed(d != NULL);
	date_return_if_failed(d->julian);
	date_return_if_failed(!d->dmy);
	date_return_if_failed(date_valid_julian(d->julian_days));

	/*
	* Formula taken from the formula was for the
	* Julian Period which starts on 1 January 4713 BC, we add
	* 1,721,425 to the number of days before doing the formula.
	* See the Calendar FAQ.
	*/

	A = d->julian_days + 1721425 + 32045;
	B = ( 4 *(A + 36524) )/ 146097 - 1;
	C = A - (146097 * B)/4;
	D = ( 4 * (C + 365) ) / 1461 - 1;
	E = C - ((1461*D) / 4);
	M = (5 * (E - 1) + 2)/153;

	m = M + 3 - (12*(M/10));
	day = E - (153*M + 2)/5;
	y = 100 * B + D - 4800 + (M/10);

	d->month = m;
	d->day   = day;
	d->year  = y;

	d->dmy = DATE_TRUE;
}

/*
 * Set and get based functions
 */

DateBoolean date_set_dmy(Date *d, DateDay day, DateMonth m, DateYear y)
{
	date_return_val_if_failed(d != NULL,DATE_FALSE);

	if (!date_valid_dmy(day, m, y)) {
		d->status = DATE_FALSE;
		return DATE_FALSE;
	}

	d->julian = DATE_FALSE;

	d->month = m;
	d->day   = day;
	d->year  = y;

	d->dmy = DATE_TRUE;
	return DATE_TRUE;
}

DateBoolean date_set_julian(Date *d, int j)
{
	date_return_val_if_failed(d != NULL, DATE_FALSE);

	if (!date_valid_julian(j)) {
		d->status = DATE_FALSE;
		return DATE_FALSE;
	}

	d->julian_days = j;
	d->julian = DATE_TRUE;
	d->dmy = DATE_FALSE;
	return DATE_TRUE;
}

DateBoolean date_set_day(Date *d, DateDay day)
{
	int index;
	date_return_val_if_failed(d != NULL, DATE_FALSE);

	d->julian = DATE_FALSE;

	index = date_is_leap_year(d->year) ? 1 : 0;
	if (day >= days_in_months[index][d->month]) {
		day = day - (days_in_months[index][d->month]);
		d->day = days_in_months[index][d->month];
		date_add_days(d,day);
		date_update_dmy(d);
	} else {
		d->day = day;
	}

	if (date_valid_dmy(d->day, d->month, d->year)) {
		d->status = d->dmy = DATE_TRUE;
	} else {
		d->status = d->dmy = DATE_FALSE;
	}
	return d->status;
}

DateBoolean date_set_month(Date *d, DateMonth month)
{
	date_return_val_if_failed(d != NULL, DATE_FALSE);
	/*date_return_val_if_failed(date_valid_month(month), DATE_FALSE);*/

	if (d->julian && !d->dmy) {
		date_update_dmy(d);
	}
	d->julian = DATE_FALSE;
	if (month>12) {
		d->month = 12;
		date_add_months(d, month-12);
	} else if (d->month==1 && month==0) {
		d->month = 12;
		date_sub_years(d, 1);
	} else {
		d->month = month;
	}
	if (date_valid_dmy(d->day, d->month, d->year)) {
		d->dmy = DATE_TRUE;
	} else {
		d->status = DATE_FALSE;
		d->dmy = DATE_FALSE;
	}
	return d->status;
}

DateBoolean date_set_year(Date *d, DateYear year)
{
	date_return_val_if_failed(d != NULL,DATE_FALSE);
	date_return_val_if_failed(date_valid_year (year),DATE_FALSE);

	if (d->julian && !d->dmy) {
		date_update_dmy(d);
	}
	d->julian = DATE_FALSE;

	d->year = year;

	if (date_valid_dmy(d->day, d->month, d->year)) {
		d->dmy = DATE_TRUE;
	} else {
		d->status = DATE_FALSE;
		d->dmy = DATE_FALSE;
	}
	return d->status;
}

/* Get Date properties */
DateDay date_get_day(Date *d)
{
	date_return_val_if_failed(d->status > 0, DATE_FALSE);
	if (!d->dmy) {
		date_update_dmy(d);
	}
	return d->day;
}

DateMonth date_get_month(Date *d)
{
	date_return_val_if_failed(d->status > 0, DATE_FALSE);
	if (!d->dmy) {
		date_update_dmy(d);
	}
	return d->month;
}

DateYear date_get_year(Date *d)
{
	date_return_val_if_failed(d->status > 0, DATE_FALSE);
	if (!d->dmy) {
		date_update_dmy(d);
	}
	return d->year;
}

int date_get_julian(Date *d)
{
	date_return_val_if_failed(d->status > 0, DATE_FALSE);
	if (!d->julian) {
		date_update_julian(d);
	}
	return d->julian_days;
}


/* Date basic operations */

DateBoolean date_add_days(Date *d, int days)
{
	date_return_val_if_failed(d->status > 0, DATE_FALSE);
	if (days < 0) {
		return date_sub_days(d,-days);
	}
	if (!d->julian) {
		date_update_julian(d);
	}
	if (days <= (DATE_JULIAN_MAX-d->julian_days)) {
		d->julian_days += days;
		d->dmy = DATE_FALSE;
		return DATE_TRUE;
	}
	return DATE_FALSE;
}

DateBoolean date_sub_days(Date *d, int days)
{
	date_return_val_if_failed(d->status > 0, DATE_FALSE);
	if (!d->julian) {
		date_update_julian(d);
	}
	if (days < 0) {
		return date_add_days(d,-days);
	}
	date_return_val_if_failed(d->julian_days>days, DATE_FALSE);
	d->julian_days -= days;
	d->dmy = DATE_FALSE;
	return DATE_TRUE;
}

DateBoolean date_add_weeks(Date *d, int weeks)
{
	date_return_val_if_failed(d->status > 0, DATE_FALSE);
	return date_add_days(d, weeks*7);
}

DateBoolean date_sub_weeks(Date *d, int weeks)
{
	date_return_val_if_failed(d->status > 0, DATE_FALSE);
	return date_sub_days(d, weeks*7);
}

DateBoolean date_add_months(Date *d, int nmonths)
{
	date_uint years, months;
	int index;

	date_return_val_if_failed(d->status > 0, DATE_FALSE);
	if (nmonths < 0) {
		return date_sub_months(d, -nmonths);
	}
	date_return_val_if_failed(d != NULL, DATE_FALSE );
	date_return_val_if_failed(date_valid(d), DATE_FALSE );

	if (!d->dmy) {
		date_update_dmy(d);
	}
	date_return_val_if_failed(d->dmy, DATE_FALSE);

	nmonths += d->month - 1;

	months = nmonths%12;
	years  = nmonths/12;

	d->month = months + 1;
	if (years <= (DateYear)DATE_YEAR_MAX-d->year) {
		d->year  += years;
	} else {
		return DATE_FALSE;
	}

	index = date_is_leap_year(d->year) ? 1 : 0;

	if (d->day > days_in_months[index][d->month]) {
		d->day = days_in_months[index][d->month];
	}

	d->julian = DATE_FALSE;

	date_return_val_if_failed(date_valid(d), DATE_FALSE);
	return DATE_TRUE;
}

DateBoolean date_sub_months(Date *d, int nmonths)
{
	date_uint years, months;
	int index;

	date_return_val_if_failed(d->status > 0,  DATE_FALSE);
	if (nmonths < 0) {
		return date_add_months(d, -nmonths);
	}
	date_return_val_if_failed(d != NULL, DATE_FALSE );
	date_return_val_if_failed(date_valid(d), DATE_FALSE );

	if (!d->dmy) {
	  date_update_dmy (d);
	}
	date_return_val_if_failed(d->dmy, DATE_FALSE);

	years  = nmonths / 12;
	months = nmonths % 12;

	date_return_val_if_failed(d->year > years, DATE_FALSE);

	if (years > d->year) {
		return DATE_FALSE;
	}
	d->year -= years;

	if (d->month > months) {
		d->month -= months;
	} else {
		months -= d->month;
		d->month = 12 - months;
		d->year -= 1;
	}

	index = date_is_leap_year(d->year) ? 1 : 0;

	if (d->day > days_in_months[index][d->month]) {
		d->day = days_in_months[index][d->month];
	}

	d->julian = DATE_FALSE;

	date_return_val_if_failed(date_valid(d),DATE_FALSE);
	return DATE_TRUE;
}

DateBoolean date_add_years(Date *d, int years)
{
	if (years < 0) {
		return date_sub_years(d, -years);
	}
	date_return_val_if_failed(d != NULL, DATE_FALSE);
	date_return_val_if_failed(date_valid(d), DATE_FALSE);

	if (!d->dmy) {
		date_update_dmy(d);
	}
	date_return_val_if_failed(d->dmy, DATE_FALSE);
	d->year += years;
	if (d->month == 2 && d->day == 29) {
		if (!date_is_leap_year(d->year)) {
			d->day = 28;
		}
	}
	d->julian = DATE_FALSE;
	date_return_val_if_failed(date_valid(d), DATE_FALSE);
	return DATE_TRUE;
}

DateBoolean date_sub_years(Date *d, int years)
{
	if (years < 0) {
		return date_add_years(d, -years);
	}
	date_return_val_if_failed(d != NULL, DATE_FALSE);
	date_return_val_if_failed(date_valid(d), DATE_FALSE);

	if (!d->dmy) {
		date_update_dmy(d);
	}
	date_return_val_if_failed(d->dmy, DATE_FALSE);
	date_return_val_if_failed(d->year > years, DATE_FALSE);

	d->year -= years;

	if (d->month == 2 && d->day == 29) {
		if (!date_is_leap_year(d->year)) {
			d->day = 28;
		}
	}
	d->julian = DATE_FALSE;
	return DATE_TRUE;
}

/*
 * Sort, Compare, Clamp dates
 *
 */

int date_compare(Date *d1, Date *d2)
{
	date_return_val_if_failed(d1 != NULL, 0);
	date_return_val_if_failed(d2 != NULL, 0);
	date_return_val_if_failed(date_valid(d1), 0);
	date_return_val_if_failed(date_valid(d2), 0);
	while (DATE_TRUE) {
		if (d1->julian && d2->julian) {
			if (d1->julian_days < d2->julian_days) {
				return -1;
			} else if (d1->julian_days > d2->julian_days) {
				return 1;
			} else {
				return 0;
			}
		} else if (d1->dmy && d2->dmy) {
			if (d1->year < d2->year) {
				return -1;
			} else if (d1->year > d2->year) {
				return 1;
			} else {
				if (d1->month < d2->month) {
					return -1;
				} else if (d1->month > d2->month) {
					return 1;
				} else {
					if (d1->day < d2->day) {
						return -1;
					} else if (d1->day > d2->day) {
						return 1;
					} else {
						return 0;
					}
				}
			}
		} else {
			if (!d1->julian) {
				date_update_julian(d1);
			}
			if (!d2->julian) {
				date_update_julian(d2);
			}
			date_return_val_if_failed(d1->julian, 0);
			date_return_val_if_failed(d2->julian, 0);
		}
	}
	return 0; /* warnings */
}

/**
 * If date is prior to min_date, sets date equal to min_date.
 * If date falls after max_date, sets date equal to max_date.
 * Either min_date and max_date may be NULL. All non-NULL dates must be valid.
 */
void date_clamp(Date *date, Date *min_date, Date *max_date)
{
	date_return_if_failed(date);
	date_return_if_failed(date_valid(date));

	if (min_date != NULL) {
		date_return_if_failed(date_valid(min_date));
	}

	if (max_date != NULL) {
		date_return_if_failed(date_valid(max_date));
	}

	/*
	 * Do not affect min_date & max_date values by swapping them
	 * Just invert the test
	 */
	if (min_date != NULL && max_date != NULL){
		if (date_compare (min_date, max_date) > 0) {
			if (min_date && date_compare(date, min_date)>0) {
				*date = *min_date;
			}
			if (max_date && date_compare(date, max_date)<0) {
				*date = *max_date;
			}
		} else {
			if (min_date && date_compare(date, min_date)<0) {
				*date = *min_date;
			}
			if (max_date && date_compare(date, max_date)>0) {
				*date = *max_date;
			}
		}
	}
}

/**
 * If date is prior to min_date, sets date equal to min_date.
 * If date falls after max_date, sets date equal to max_date.
 * Either min_date and max_date may be NULL. All non-NULL dates must be valid.
 */
void date_order(Date *d1, Date *d2)
{
	date_return_if_failed(d1 != NULL);
	date_return_if_failed(d2 != NULL);
	date_return_if_failed(date_valid(d1));
	date_return_if_failed(date_valid(d2));

	if (date_compare(d1, d2) > 0) {
		Date tmp = *d1;
		*d1 = *d2;
		*d2 = tmp;
	}
}

/* userfriendly & misc functions */

DateBoolean date_is_first_of_month(Date *d)
{
	date_return_val_if_failed(d != NULL, DATE_FALSE);
	date_return_val_if_failed(date_valid(d), DATE_FALSE);

	if (!d->dmy) {
		date_update_dmy(d);
	}
	date_return_val_if_failed(d->dmy, DATE_FALSE);

	return d->day == 1 ? DATE_TRUE : DATE_FALSE;
}

DateBoolean date_is_last_of_month(Date *d)
{
	int index;

	date_return_val_if_failed(d != NULL, DATE_FALSE);
	date_return_val_if_failed(date_valid(d), DATE_FALSE);

	if (!d->dmy) {
		date_update_dmy (d);
	}
	date_return_val_if_failed(d->dmy, DATE_FALSE);

	index = date_is_leap_year(d->year) ? 1 : 0;

	if (d->day == days_in_months[index][d->month]) {
		return DATE_TRUE;
	} else {
		return DATE_FALSE;
	}
}

DateWeekday date_get_weekday(Date *d)
{
	date_return_val_if_failed(d != NULL, DATE_FALSE);
	if (!d->julian) {
		date_update_julian(d);
	}

	date_return_val_if_failed(d->julian, DATE_BAD_WEEKDAY);

	return ((d->julian_days - 1) % 7) + 1;
}

int date_get_day_of_year(Date *d)
{
	int index;

	date_return_val_if_failed(d != NULL, 0);

	if (!d->dmy) {
		date_update_dmy(d);
	}

	date_return_val_if_failed(date_valid(d), 0);

	date_return_val_if_failed(d->dmy, 0);

	index = date_is_leap_year(d->year) ? 1 : 0;

	return (days_in_year[index][d->month] + d->day);
}

int date_get_week_of_year(Date *d)
{
	DateWeekday wd;
	int day;
	Date first;

	date_return_val_if_failed(d != NULL, 0);
	date_return_val_if_failed(date_valid(d), 0);

	if (!d->dmy) {
		date_update_dmy (d);
	}
	date_return_val_if_failed(d->dmy, 0);

	date_clear(&first, 1);
	date_set_dmy(&first, 1, 1, d->year);

	if (d->weekstart == DATE_STARTWEEK_MONDAY) {
		wd = date_get_weekday(&first) - 1;
		day = date_get_day_of_year(d) - 1;
	} else {
		wd = date_get_weekday(&first);
		if (wd == 7) {
			wd = 0;
		}
		day = date_get_day_of_year(d) - 1;
	}
	return ((DateDay)((day + wd)/7U) + (wd == 0 ? 1 : 0));
}

/* 'A la' Date::Calc
 * Still to define which I have to use
 */
int date_get_week_of_year2(Date *d)
{
	DateDay first;
	Date d0101;

	date_return_val_if_failed(d != NULL, 0);
	date_return_val_if_failed(date_valid(d), 0);

	if (!d->dmy) {
		date_update_dmy(d);
	}
	date_return_val_if_failed(d->dmy, 0);

	d0101.weekstart = d->weekstart;
	date_set_dmy(&d0101, 1, 1, d->year);
	first  = date_get_weekday(&d0101)-1;

	return( (int)( (date_days_between(&d0101, d) + first) / (int)7 ) + (first < 4) );
}

/*
 * Implementaion of this algorithm:
 * http://personal.ecu.edu/mccartyr/ISOwdALG.txt
 */
int date_get_isoweek_of_year(Date *date)
{
	int y,m,d;
	int yy, c, g, h, i;
	int jan1_weekday;
	int yearnumber, weeknumber = 1, weekday;
	/* [0] is only an empty cell */
	int mnth[13] = {0,0,31,59,90,120,151,181,212,243,273,304,334};
	int DayOfYearNumber;

	DateBoolean y_1_isleap;
	DateBoolean y_isleap;

	date_return_val_if_failed(date != NULL, 0);
	date_return_val_if_failed(date_valid(date), 0);

	if (!date->dmy) {
		date_update_dmy(date);
	}
	date_return_val_if_failed(date->dmy, 0);

	m	= date->month;
	y	= date->year;
	d	= date->day;

	y_isleap = date_is_leap_year(y);
	y_1_isleap = date_is_leap_year(y - 1);

	DayOfYearNumber = d + mnth[m];

	if(y_isleap && m > 2){
		DayOfYearNumber++;
	}

	yy = (y-1) % 100;
	c = (y-1) - yy;
	g = yy + (int)ceil(yy / 4);
	jan1_weekday = 1 + (int)((((((c / 100) % 4) * 5) + g) % 7));

	/* weekday for year-month-day */
	h = DayOfYearNumber + (jan1_weekday - 1);
	weekday = 1 + ((h - 1) % 7);

	/* find if Y M D falls in YearNumber Y-1, WeekNumber 52 or 53 */
	if (DayOfYearNumber <= (8 - jan1_weekday) && jan1_weekday > 4){
		yearnumber = y - 1;
		if (jan1_weekday == 5 || (jan1_weekday == 6 && y_1_isleap)) {
			weeknumber = 53;
		} else {
			weeknumber = 52;
		}
	} else {
		yearnumber = y;
	}
	/* find if Y M D falls in YearNumber Y+1, WeekNumber 1 */
	if (yearnumber == y) {
		if (y_isleap) {
			i = 366;
		} else {
			i = 365;
		}
		if ((i - DayOfYearNumber) < (4 - weekday)) {
			yearnumber++;
		}
	}
	/* find if Y M D falls in YearNumber Y, WeekNumber 1 through 53 */
	if (yearnumber == y) {
		i = DayOfYearNumber + (7 - weekday) + (jan1_weekday - 1);
		weeknumber = (int)ceil(i / 7); /* kludge!!! - JMC */
		if (jan1_weekday > 4) {
			weeknumber--;
		}
	}
	return weeknumber;
}


int date_get_weeks_of_year(DateYear year)
{
	Date d;
	int ws;

	date_return_val_if_failed(date_valid_year (year), 0);

	ws = d.weekstart;
	date_clear(&d, 1);
	date_set_dmy(&d, 1, 1, year);

	if (date_get_weekday(&d) == ws) {
		return 53;
	}
	date_set_dmy (&d, 31, 12, year);
	if (date_get_weekday(&d) == ws) {
		return 53;
	}

	if (date_is_leap_year(year)) {
		date_set_dmy(&d, 2, 1, year);
		if (date_get_weekday(&d) == ws) {
			return 53;
		}
		date_set_dmy(&d, 30, 12, year);
		if (date_get_weekday(&d) == ws) {
			return 53;
		}
	}
	return 52;
}


/**
 * Computes the number of days between two dates.
 * If date2 is prior to date1, the returned value is negative. Both dates must be valid.
 */
DateDay date_days_between(Date *d1, Date *d2)
{
	date_return_val_if_failed(d1 != NULL, 0);
	date_return_val_if_failed(d2 != NULL, 0);

	date_return_val_if_failed(date_valid(d1), 0);
	date_return_val_if_failed(date_valid(d2), 0);

	return (int)date_get_julian(d2) - (int)date_get_julian(d1);
}

DateBoolean date_set_to_weekday(Date *d, DateWeekday weekday, int n)
{
	DateWeekday wd;
	int diff;
	DateMonth m;
	DateDay day;
	date_return_val_if_failed(d != NULL, DATE_FALSE);
	date_return_val_if_failed(date_valid_weekday(weekday), DATE_FALSE);
	date_return_val_if_failed(date_valid(d), DATE_FALSE);

	m = date_get_month(d);
	day = date_get_day(d);
	if (n>0) {
		/* count from the start of d->month */
		date_set_day(d, 1);
		wd = date_get_weekday(d);
		diff = weekday - wd;
		if (diff < 0) {
			diff += 7;
		}
		diff += 7*(n - 1);
		date_add_days(d, diff);
	} else {
		/* count from the end of d->month */
		date_set_to_lastmonthday(d);
		wd = date_get_weekday(d);
		diff = wd - weekday;
		if (diff < 0) {
			diff += 7;
		}
		diff += 7 * (-n-1);
		date_sub_days(d, diff);
	}
	if (m != date_get_month(d)) {
		date_set_dmy(d, day, m, date_get_year(d));
		return DATE_FALSE;
	}
	return DATE_TRUE;
}

Date *date_get_to_weekday(Date *d, DateWeekday weekday, int n)
{
	Date *newd;

	date_return_val_if_failed(d != NULL, DATE_FALSE);
	date_return_val_if_failed(date_valid_weekday(weekday), DATE_FALSE);
	date_return_val_if_failed(date_valid(d), DATE_FALSE);

	newd = date_new(d->weekstart);
	date_return_val_if_failed(newd != NULL, DATE_FALSE);

	date_set_julian(newd, date_get_julian(d));

	if (!date_set_to_weekday(newd, weekday, n)) {
		return NULL;
	}
	return newd;
}


DateBoolean date_set_to_lastmonthday(Date *d)
{
	return date_set_day(d, date_is_leap_year(d->year) ? days_in_months[1][d->month] : days_in_months[0][d->month]);
}

Date* date_get_to_lastmonthday(Date *d)
{
	Date *newd;

	date_return_val_if_failed(d != NULL, DATE_FALSE);
	date_return_val_if_failed(date_valid(d), DATE_FALSE);

	newd = date_new(d->weekstart);
	date_return_val_if_failed(newd != NULL, DATE_FALSE);
	date_set_julian(newd, date_get_julian(d));

	if (!date_set_to_lastmonthday(newd)) {
		return NULL;
	}
	return newd;
}


DateBoolean date_set_first_dow(Date *d)
{
	DateWeekday wd;
	date_return_val_if_failed(d != NULL, DATE_FALSE);
	date_return_val_if_failed(date_valid(d), DATE_FALSE);

	wd = date_get_weekday(d);
	if (d->weekstart == DATE_STARTWEEK_MONDAY) {
		if (wd!=DATE_MONDAY) {
			return date_sub_days(d, wd-DATE_MONDAY);
		}
	} else {
		if (wd!=DATE_SUNDAY) {
			return date_sub_days(d, wd);
		}
	}
	return DATE_TRUE;
}

Date* date_get_first_dow(Date *d)
{
	Date *newd;

	date_return_val_if_failed(d != NULL, DATE_FALSE);
	date_return_val_if_failed(date_valid(d), DATE_FALSE);

	newd = date_new(d->weekstart);
	date_return_val_if_failed(newd != NULL, DATE_FALSE);
	date_set_julian(newd, date_get_julian(d));
	if (date_set_first_dow(newd)) {
		date_free(newd);
		return NULL;
	}
	return newd;
}

DateBoolean date_set_last_dow(Date *d)
{
	DateWeekday wd;
	date_return_val_if_failed(d != NULL, DATE_FALSE);
	date_return_val_if_failed(date_valid(d), DATE_FALSE);

	wd = date_get_weekday(d);
	if (d->weekstart == DATE_STARTWEEK_MONDAY) {
		if (wd != DATE_SUNDAY) {
			return date_add_days(d, DATE_SUNDAY - wd);
		}
	} else {
		if (wd != DATE_SATURDAY) {
			return date_add_days(d, DATE_SATURDAY - wd);
		}
	}
	return DATE_TRUE;
}

Date* date_get_last_dow(Date *d)
{
	Date *newd;

	date_return_val_if_failed(d != NULL, DATE_FALSE);
	date_return_val_if_failed(date_valid(d), DATE_FALSE);

	newd = date_new(d->weekstart);
	date_return_val_if_failed(newd != NULL, DATE_FALSE);
	date_set_julian(newd, date_get_julian(d));
	if (!date_set_last_dow(newd)) {
		date_free(newd);
		return NULL;
	}
	return newd;
}

DateBoolean date_settoweekdayinsameweek(Date *d, DateWeekday wd)
{
	DateWeekday currentwd;

	date_return_val_if_failed(d != NULL, DATE_FALSE);
	date_return_val_if_failed(date_valid(d), DATE_FALSE);
	date_return_val_if_failed(date_valid_weekday(wd), DATE_FALSE);

	currentwd = date_get_weekday(d);
	if (currentwd != wd) {
		if (wd < currentwd) {
			return date_sub_days(d, currentwd-wd);
		} else {
			return date_add_days(d, wd-currentwd);
		}
	}
	return DATE_TRUE;
}

Date* date_gettoweekdayinsameweek(Date *d, DateWeekday wd)
{
	Date *newd;

	date_return_val_if_failed(d != NULL, DATE_FALSE);
	date_return_val_if_failed(date_valid(d), DATE_FALSE);
	date_return_val_if_failed(date_valid_weekday(wd), DATE_FALSE);

	newd = date_new(d->weekstart);
	date_return_val_if_failed(newd != NULL, DATE_FALSE);
	date_set_julian(newd, date_get_julian(d));
	if (!date_settoweekdayinsameweek(newd, wd)) {
		date_free(newd);
		return NULL;
	}
	return newd;
}

DateBoolean date_settonextweekday(Date *d, DateWeekday wd)
{
	DateWeekday currentwd;
	int diff;

	date_return_val_if_failed(d != NULL, DATE_FALSE);
	date_return_val_if_failed(date_valid(d), DATE_FALSE);
	date_return_val_if_failed(date_valid_weekday(wd), DATE_FALSE);

	currentwd = date_get_weekday(d);
	if (currentwd == wd) {
		return DATE_TRUE;
	}
	if (wd < currentwd) {
		diff = 7 - (currentwd - wd);
	} else {
		diff = wd - currentwd;
	}
	if (!date_add_days(d, diff)) {
		return DATE_FALSE;
	}
	return DATE_TRUE;
}

Date* date_gettonextweekday(Date *d, DateWeekday wd)
{
	Date *newd;

	newd = date_new(d->weekstart);
	date_return_val_if_failed(newd != NULL, DATE_FALSE);

	if (!date_settonextweekday(newd, wd)) {
		date_free(newd);
		return NULL;
	}
	return newd;
}


DateBoolean date_settoprevweekday(Date *d, DateWeekday wd)
{
	DateWeekday currentwd;
	int diff;

	date_return_val_if_failed(d != NULL, DATE_FALSE);
	date_return_val_if_failed(date_valid(d), DATE_FALSE);
	date_return_val_if_failed(date_valid_weekday(wd), DATE_FALSE);

	currentwd = date_get_weekday(d);
	if (wd == currentwd) {
		return DATE_TRUE;
	}
	if (wd > currentwd) {
		diff = 7 - (wd - currentwd);
	} else {
		diff = currentwd - wd;
	}
	if (!date_sub_days(d, diff)) {
		return DATE_FALSE;
	}
	return DATE_TRUE;
}

Date* date_gettoprevweekday(Date *d, DateWeekday wd)
{
	Date *newd;

	newd = date_new(d->weekstart);
	date_return_val_if_failed (newd != NULL, DATE_FALSE);
	if (!date_settoprevweekday(newd, wd)) {
		date_free(newd);
		return NULL;
	}
	return newd;
}

/* Returns the number of days in a given month of a given year */
DateBoolean date_settotheweek(Date *d, DateWeek week, DateWeekday wd)
{
	int year;

	date_return_val_if_failed(d != NULL, DATE_FALSE);
	date_return_val_if_failed(date_valid(d), DATE_FALSE);
	date_return_val_if_failed(date_valid_weekday(wd), DATE_FALSE);

	year = d->year;

	/* Jan 4 always lies in the 1st week of the year */
	date_set_dmy(d, 4, DATE_JANUARY, year);
	if (!date_settoweekdayinsameweek(d, wd)) {
		return DATE_FALSE;
	} else {
		date_add_weeks(d, week - 1);
	}
	if (date_get_year(d) != year) {
		return DATE_FALSE;
	}
	return DATE_TRUE;
}

/* Returns the number of days in a given month of a given year */
Date *date_gettotheweek(Date *d, DateWeek week, DateWeekday wd)
{
	Date *newd;

	date_return_val_if_failed(d != NULL, DATE_FALSE);
	date_return_val_if_failed(date_valid(d), DATE_FALSE);
	date_return_val_if_failed(date_valid_weekday(wd), DATE_FALSE);

	newd = date_new_dmy(d->day,d->month,d->year,d->weekstart);
	if (newd && date_settotheweek(newd, week, wd)) {
		return newd;
	} else {
		return NULL;
	}

}

/* Returns the number of days in a given month of a given year */
int date_get_days_in_month(DateMonth month, DateYear year)
{
	int index;
	index = date_is_leap_year(year) ? 1 : 0;
	return days_in_months[index][month];

}

Date *date_get_easter(DateYear year, DateDay weekstart)
{
	int golden, solar, lunar, pfm, dom, tmp, easter;
	Date *date;

	/* the Golden number */
	golden = (year % 19) + 1;

	if (year <= 1752) {
		/* JULIAN CALENDAR */
		/* the "Dominical number" - finding a Sunday */
		dom = (year + (year / 4) + 5) % 7;
		if (dom < 0) {
			dom += 7;
		}

		/* uncorrected date of the Paschal full moon */
		pfm = (3 - (11 * golden) - 7) % 30;
		if (pfm < 0) {
			pfm += 30;
		}
	} else {
		/* GREGORIAN CALENDAR */
		/* the "Dominical number" - finding a Sunday */
		dom = (year + (year / 4) - (year / 100) + (year / 400)) % 7;
		if (dom < 0) {
			dom += 7;
		}

		/* the solar and lunar corrections */
		solar = (year - 1600) / 100 - (year - 1600) / 400;
		lunar = (((year - 1400) / 100) * 8) / 25;

		/* uncorrected date of the Paschal full moon */
		pfm = (3 - (11*golden) + solar - lunar) % 30;
		if (pfm < 0) {
			pfm += 30;
		}
	}

	/* corrected date of the Paschal full moon - days after 21st March */
	if ((pfm == 29) || (pfm == 28 && golden > 11)) {
		pfm--;
	}

	tmp = (4-pfm-dom) % 7;
	if (tmp < 0) {
		tmp += 7;
	}

	/* Easter as the number of days after 21st March */
	easter = pfm + tmp + 1;

	if (easter < 11) {
		date = date_new_dmy(easter + 21, 3, year, weekstart);
	} else {
		date = date_new_dmy(easter - 10, 4, year, weekstart);
	}
	return date;
}
